; This programe is a Boot Sector Loader written by Jemin Z. M. Chen.
; on 2012/03/13. Please use it under the GNU GPL2 copyright.
;
; It contains the cloned BPB, the partition table and the loader.
; The cloned BPB is used to formulate as the image is a FDD.
; The partition table is used for the image working as a normal hard disk.
; The loader will detect where the image is booted from, a hdd or a fdd
; then it prepares the boot sector with proper value in the cloned BPB, so 
; the system is ready to boot up.
;
; Please mount it as a hard disk when reformatting, then refer to the 
; real BPB, fix the fields of the cloned BPB before it be mounted as 
; a floppy.
;
; Compatibility: FAT12/FAT16 system ( tested: W95b, W98, FreeDOS ke2041)
;                Dos6.22/W95A unable to boot up as a floppy, mount neither.
;
; I've build a 4M image with freedos in the attachment for the one who
; is willing to have a trial. Using this MBR sector and the patch I posted,
; we have 6 mount points on DOSBox.
;
; Changes:
; 12/03/20	Fix error setting of DS after shadowing
;		Add more codes to read remaining sectors
;		But there is lack of space to put the FAT32 structure
;		Have to build the FAT32 header in our cloned BPB

.model MEDIUM, C
.386

MBR_SIZE= _SECTOR_END-_0000

BOOT_POS= 7C00h

JMP_CODE_FAR= 0EAh
RET_CODE_FAR= 0CBh


JMP_CODE= 0EBh
JMP_SHORT_BYTES= 2
JMP_CODE_OFFSET= (0FFh AND ( TYPE Bios_Parameter_Block - JMP_SHORT_BYTES))
NOP_CODE= 090h

PARAMETER_SPACE= 8
bFileSystemID_BP TEXTEQU < Byte PTR [bp-8]>
BootDriveNo TEXTEQU < Byte PTR [bp-6]>
STARTING_SECTOR_D TEXTEQU < [bp-4]>
STARTING_SECTOR_LO TEXTEQU < [bp-4]>
STARTING_SECTOR_HI TEXTEQU < [bp-2]>

INT_VECTOR_1Eh= 1Eh shl 2
LENGTH_INT_1Eh= 11
OFFSET_new_int_1eh= 522h
OFFSET_int_1eh_SecPerTrk= 4
OFFSET_int_1eh_SpinTime= 9

TwoWords STRUC
	LoWord	dW	?
	HiWord	dW	?
TwoWords ENDS

DWORDS	UNION
	DWORD	?
	TwoWords	<>
DWORDS	ENDS

Partition_Table	STRUC
	bBootable	dB	?	;0
	bStart_Side	dB	?	;1
	bStart_Sec	dB	?	;2
	bStart_Cyl	dB	?	;3
	bSystemID	dB	?	;4	1= DOS-12, 4= DOS-16		
	bEnd_Side	dB	?	;5
	bEnd_Sector	dB	?	;6
	bEnd_Cyl	dB	?	;7
	dRelat_Sec	DWORDS	<>	;8	OK mov EDX, part1.dRelat_Sec
	dTotal_Sec	DWORDS	<>	;C
Partition_Table	ENDS


Bios_Parameter_Block STRUC
	bEntry_Code0	DB	JMP_CODE	;00
	bEntry_Code1	DB	JMP_CODE_OFFSET	;01
	bEntry_Code2	DB	NOP_CODE	;02
	bsVenderID	DB	8 dup( " ")	;03
	wBytePerSec	DW	?		;0B
	bSecPerClu	DB	?		;0D
	wRevSec		DW	?		;0E
	bNumOfFAT	DB	?		;10
	wEntOfRtDir	DW	?		;11
	wTotSec		DW	?		;13
	bMediaDscr	DB	?		;15
	wSecPerFAT	DW	?		;16
	wSecPerTrk	DW	?		;18
	wNumOfHd	DW	?		;1A
	dNumOfHidSec 	DWORDS	<>		;1C
	dHugeTotSec 	DWORDS	<>		;20
	bDrvNum		DB	?		;24
	bReserved	DB	?		;25
	bSignature	DB	?		;26
	dSerialNO	DD	?		;27
	bsVol_Label	DB	11 dup( " ")	;2B
	bsFS_TYPE	DB	8 dup( " ")	;36
Bios_Parameter_Block ENDS


Dir_Entry STRUC
	bEntryName	DB	11 dup( " ");	;00
	bAttrib		DB	?		;0B
	bNTRes		DB	?		;0C
	bMilliSecondStamp	DB	?	;0D
	wCrtTime	DW	?		;0E
	wCrtDate	DW	?		;10
	wAccessDate	DW	?		;12
	wHiFirstClust	DW	?		;14
	wModTime	DW	?		;16
	wModDate	DW	?		;18
	wLoFirstClust	DW	?		;1A
	dEntrySize	DD	?		;1C
Dir_Entry ENDS


;include CHECK_IN.MAC
;include COMMON.MAC

;.code

; CS= 0000h
; IP= 7C00h
_TEXT	SEGMENT PARA	public 'CODE' ;at 07C0h
		ASSUME	CS: _TEXT, DS: _TEXT, ES: NOTHING, SS: NOTHING

ORG	0000h
_0000:
PUBLIC CODE_ORG_START
CODE_ORG_START:

BPB	Bios_Parameter_Block { \
,,,\;JumpCode
"ZM120320",\
512,\
1,\;bS/Clu
2,\;wRevS
2, \;FAT
512,\;wDir
8192,\;wTot
0F8h,\;bMedia
32,\;wS/FAT
1,\;wS/Trk
64,\;wHd
<0>,\;dHidS
<0>,\;dTotS
80h,\;bDL
0,\;bRev
41,\;bSig
4194304,\;dSN
"4M_FDOS    ",\
"FAT16   "\
}

ORG	0000h
	JMP	_START
;	JMP	_show_regs

ORG	TYPE Bios_Parameter_Block 

_START:

;_003Eh:
XOR     CX, CX
CLI

MOV     SS, CX
LEA	SP, [ BX- PARAMETER_SPACE]
;MOV	SP, BX
;SUB	SP, PARAMETER_SPACE

PUSH    CX
POP     ES			; ES= 0
MOV     BP, INT_VECTOR_1Eh	; 0078
LDS     SI, [BP]		; DS:SI= F000:1060h

PUSH    DS
PUSH    SI
PUSH    SS
PUSH    BP
MOV     DI, OFFSET_new_int_1eh	; 0522
MOV     [BP], DI		; save the new INT 1Eh vector= 0:522h
MOV     [BP+02], CX		; Diskette Parameter Table
MOV     CL, LENGTH_INT_1Eh	; 0Bh
CLD
REP MOVSB			; copy 11 Byte from old Int_1E to new
				; FE 38 03 00 CF 00 00 00 00 00 00
				; FE 38 03 02 24 00 00 00 00 F0 00


PUSH    ES
POP     DS		; DS= 0
;needn't ;MOV  BYTE PTR [ DI- LENGTH_INT_1Eh+ OFFSET_int_1eh_SpinTime], 0Fh; [052D-2]=0Fh

sti

;MOV     BP, BOOT_POS	;7C00h
MOV     BP, BX	;7C00h
MOV  BootDriveNo, DL
MOV  ( Bios_Parameter_Block PTR [BP]).bDrvNum, DL



CMP  DL, CL
JL     _HDD			; bDrvNum=00h jumped

_FDD:
MOV	AX, ( Bios_Parameter_Block PTR [BP]).wSecPerTrk ;[7C18]= 24h
; it should be AH= 0 for floppy
MOV  BYTE PTR [ DI- LENGTH_INT_1Eh+ OFFSET_int_1eh_SecPerTrk], AL; [052D-2]=0Fh

_HDD:

_CheckBootable:
;LEA	BX, [ BP+ OFFSET part1]
MOV	BX, OFFSET part1+ BOOT_POS
;MOV	BX, OFFSET part1.bBootable+ BOOT_POS
;ADD     BX, OFFSET part1.bBootable

_nextEntry:
CMP     BYTE PTR [BX+ Partition_Table.bBootable], 80h	;Bootable?
JNE     _nonBootable


_FoundBootable:
; check only one legal bootable

_CheckSystemID:

mov	CL, COUNT_FileSystemID_to_Check

MOV     AH, [BX+ Partition_Table.bSystemID]
MOV	SI, OFFSET _FileSystemID+ BOOT_POS
;LEA	SI, [BP+ OFFSET _FileSystemID]

@@:
lodsb
CMP     AL, AH

; MOVcc if SystemID= 0Bh, 0Ch, 0Eh
JNE	_not_INT13x


_check_INT13x:
push	bx
MOV     BX, 55AAh
MOV     AH, 41h		; Check Extensions Present
INT     13h
pop	bx
JC	_not_INT13x
MOV	bFileSystemID_BP, FSID_BGDOSx


_not_INT13x:
LOOP @B


; needn't
;MOV	SI, _SHADOW_Remaining+ BOOT_POS
MOV	DI, SI

; For reserving 64*512 bytes (at least 32)
MOV	AX, ((( _SECTOR_END- _0000) SHL 6) SHR 4)
MOV	ES, AX
MOV	CX, _SECTOR_END- _SHADOW_Remaining
push	es	; the destiny
push	di
REP	MOVSB
db	RET_CODE_FAR


_nonBootable:
ADD	BL, TYPE Partition_Table	; 10h to next Partition record
JNC     _nextEntry
;JMP	_showSysErr
INT     18h	; traditionally jumped to an implementation of BASIC in ROM.

;01h	FAT12
;04h	FAT16
;05h	EXTEND
;06h	BIGDOS
;07h	HPFS
;0Bh	FAT32
;0Ch	FAT32x
;0Eh	BGDOSx
;0Fh	EXTNDx

FSID_BGDOSx= 0Eh	
_FileSystemID:
db	FSID_BGDOSx, 0Bh, 0Ch
COUNT_FileSystemID_to_Check= $- _FileSystemID



_SHADOW_Remaining:

_Reload:

; 12 bytes
MOV     AX, ( DWORDS PTR [BX+Partition_Table.dRelat_Sec]).LoWord
MOV     DX, ( DWORDS PTR [BX+Partition_Table.dRelat_Sec]).HiWord
MOV	STARTING_SECTOR_LO, AX
MOV	STARTING_SECTOR_HI, DX

; 11 bytes
;MOV     EAX, [BX+ Partition_Table.dRelat_Sec]
;MOV     STARTING_SECTOR_D, EAX
;MOV	DX, STARTING_SECTOR_HI

PUSH	ES
POP	DS
MOV	ES, CX

CALL    _ReadSector_CX
JC      _ShowDiskErr


_CheckGoodSector:
cmp	WORD PTR es:[ BX- 2], 0AA55h
;cmp	WORD PTR [ bp+ MBR_SIZE- 2], 0AA55h
JNE	_ShowSysErr

MOV  DL, BootDriveNo
CMP  ( Bios_Parameter_Block PTR [BP]).bDrvNum, DL
JE  _JumpBootRecord

_Modify:
MOV	( Bios_Parameter_Block PTR [BP]).bDrvNum, DL
CMP	CX, STARTING_SECTOR_HI	; BPB.wRevSec limitation
JNE	_ShowSysErr

MOV	( Bios_Parameter_Block PTR [BP]).dNumOfHidSec, ECX
MOV	CX, ( Bios_Parameter_Block PTR [BP]).wRevSec

MOV	AX, STARTING_SECTOR_LO
ADD	( Bios_Parameter_Block PTR [BP]).wRevSec, AX
JC	_ShowSysErr

; read remaining reserved sectors
; for other FileSystem. Work? to be checked
MOV	BX, BP
CALL    _ReadSector_add_for_Next
JC      _ShowDiskErr


_JumpBootRecord:
mov	bx, bp
;xchg	bx, bp
mul	cx
MOV	DL, BootDriveNo

POP     SI
POP     DS
POP     DWORD PTR [SI]
push	cx
push	bp
inc	cx
db	RET_CODE_FAR ;retf


_ShowSysErr:
MOV     SI, offset ptrMsgSysErr+ BOOT_POS	; Invalid system disk

_show_msg:
LODSB
CBW
ADD     SI, AX

_next_char:
LODSB
CBW
INC     AX
JZ      _ShowChangeDisk

DEC     AX
JZ      _tryINT19

MOV     BX,0007h
MOV     AH,0Eh
INT     10h

JMP     _next_char

_ShowChangeDisk:
MOV     SI, offset ptrMsgChangeDisk+ BOOT_POS	; msg Replace the disk, 
JMP     _show_msg

_ShowDiskErr:
MOV     SI, offset ptrMsgDiskErr+ BOOT_POS	; msg Disk I/O error
JMP     _show_msg

_tryINT19:
INT     16h

POP     SI
POP     DS
POP     DWORD PTR [SI]
INT     19h			;  load the operating system

@@:

ptrMsgSysErr:
DB _MsgSysErr- $- 1

ptrMsgDiskErr:
DB _MsgDiskErr- $- 1

ptrMsgChangeDisk:
DB _MsgChangeDisk- $- 1


_MsgChangeDisk:
DB	0Dh, 0Ah
DB	"Wait another disk", 0Dh, 0Ah, 0

_MsgDiskErr:
DB	0Dh, 0Ah
DB	"Disk I/O Error", 0FFh

_MsgSysErr:
DB	0Dh, 0Ah
DB	"System Disk Error", 0FFh

@@:


_ReadSector_CX:
INC     CX

_ReadSector_BX:
MOV     BX, BP	;7C00h

_ReadSector:
PUSHA
; ax=   0
; cx=   1
; dx=   0
; bx= 700h
; sp=7BE2h
; bp=7C00h
; si=106Bh
; di= 52Dh


;push 8 words as the parameters for Int 13h
PUSHD	0
PUSH    DX
PUSH    AX		; AX= 0
PUSH    ES
PUSH    BX
PUSH	1
PUSH	10h


;MOV	SI, ( Bios_Parameter_Block PTR [BP]).wSecPerTrk ;[7C18]= 24h

; Input
; DX:AX dividend
; SI	divisor
; Output
; DX:AX Quotient
; CX	remainder
XCHG	CX, AX		; save AX in CX in case of large Quotient
XCHG	AX, DX		; mov DX, AX ;dividend's HiWord first
XOR	DX, DX		;
;DIV	SI
DIV ( Bios_Parameter_Block PTR [BP]).wSecPerTrk ;[7C18]= 24h
XCHG	CX, AX		; CX= the HiWord of Quotient, AX= LoWord
			; DX= the remainder of dividend's HiWord 
;DIV	SI
DIV ( Bios_Parameter_Block PTR [BP]).wSecPerTrk ;[7C18]= 24h
XCHG	CX, DX		; Let DX= the HiWord of Quotient, CX= Sector,Remainder

; now for Head and Trk
DIV	( Bios_Parameter_Block PTR [BP]).wNumOfHd ;7C1A

MOV     DH, DL		; move R, the Head, to DH
MOV	CH, AL		; move Q, the Cylinder, to CH
ROR	AH, 2
OR	CL, AH
INC	CX		; 1 <= Sector number <= 63

; INT 0x13 Function 02h
;	 AL= # of Sector to read
;	 ES:BX buffer address
;	 CL[7:6]CH[7:0]= track
;	 CL[5:0]= Sector, 1 to 63
;	 DH= Head
;	 DL= Drive

MOV     AX, 0201h	; int 13h, read 1 sector

CMP	bFileSystemID_BP, FSID_BGDOSx

JNZ     @F	;jumped

MOV     AH, 42h
MOV     SI, SP

@@:
MOV	DL, ( Bios_Parameter_Block PTR [BP]).bDrvNum ; [7C24]= 80h
INT     13h


;  BIOS call "INT 0x13 Function 0x42" to read sectors from disk into memory
; 	Call with	%ah = 0x42
; 			%dl = drive number
; 			%ds:%si = segment:offset of disk address packet
; 	Return:
; 			%al = 0x0 on success; err code on failure

POPA
POPA		; AX= 0
JC      _ReadSector_RET	; no jump


_ReadSector_add_for_Next:
ADD	BX, ( Bios_Parameter_Block PTR [BP]).wBytePerSec ;7C0B
INC     AX
ADC	DX, 0

; cx= 1

LOOP     _ReadSector

_ReadSector_RET:

RET


@@:

ORG 01BEh
_01BE:

part1 Partition_Table{\
80h,\;bBootable
1,\;bStart_Side
1,\;bStart_Sec
0,\;bStart_Cyl
04h,\;bSystemID
63,\;bEnd_Side
1,\;bEnd_Sector
127,\;bEnd_Cyl
<1>,\;dRelat_Sec
<8191>\;dTotal_Sec
}

ORG 01D2h
db	0

ORG 01E2h
db	0

ORG 01F2h
db	0

ORG	1FEh
DB	55h, 0AAh

_SECTOR_END:

_TEXT	ENDS


	END

